from typing import Tuple
from nepali_calendar_utils.calendar_model.nepali_calendar_defaults import *
from nepali_calendar_utils.data.nepali_year_month_map import *
from nepali_calendar_utils.data.custom_calendar import *
from datetime import date

class DateConverters:
    min_nepali_year = NepaliCalendarDefaults.NepaliYearRange[0]
    max_nepali_year = NepaliCalendarDefaults.NepaliYearRange[-1]
    min_english_year = NepaliCalendarDefaults.EnglishYearRange[0]
    max_english_year = NepaliCalendarDefaults.EnglishYearRange[-1]

    @staticmethod
    def get_total_days_in_nepali_month(nepaliYYYY: int, nepaliMM: int) -> int:
        return days_in_month_map[nepaliYYYY][nepaliMM]

    @staticmethod
    def convert_to_nepali_calendar(englishYYYY: int, englishMM: int, englishDD: int) -> CustomCalendar:
        if not DateConverters.is_english_date_in_conversion_range(englishYYYY, englishMM, englishDD):
            raise ValueError(f"Out of Range: English year {englishYYYY} is out of range to convert.")

        starting_english_date, starting_nepali_calendar = DateConverters.initialize_starting_dates(
            englishYYYY, False
        )

        english_base_date = date(
            year=starting_english_date.year,
            month=starting_english_date.month,
            day=starting_english_date.day_of_month
        )

        target_english_date = date(
            year=englishYYYY,
            month=englishMM,
            day=englishDD
        )

        total_days_difference = (target_english_date - english_base_date).days

        nepaliYYYY, nepaliMM, nepaliDD = (
            starting_nepali_calendar.year,
            starting_nepali_calendar.month,
            starting_nepali_calendar.day_of_month
        )
        day_of_week = starting_nepali_calendar.day_of_week

        day_of_year = starting_nepali_calendar.day_of_year
        week_of_year = starting_nepali_calendar.week_of_year
        week_of_month = starting_nepali_calendar.week_of_month

        first_day_of_month = starting_nepali_calendar.first_day_of_month
        last_day_of_month = starting_nepali_calendar.last_day_of_month

        total_days_in_month = days_in_month_map[nepaliYYYY][nepaliMM]

        for _ in range(total_days_difference):
            nepaliDD += 1
            day_of_year += 1
            day_of_week += 1
            
            if day_of_week > 7:
                day_of_week = 1
                week_of_year += 1
                week_of_month += 1

            if nepaliDD > total_days_in_month:
                nepaliMM += 1
                nepaliDD = 1

                if nepaliMM > 12:
                    nepaliYYYY += 1
                    nepaliMM = 1
                    day_of_year = 1
                    week_of_year = 1

                week_of_month = 1
                total_days_in_month = days_in_month_map[nepaliYYYY][nepaliMM]
                first_day_of_month = day_of_week

            remaining_days = total_days_in_month - nepaliDD
            last_day_of_month = (day_of_week + remaining_days) % 7 or 7

        return CustomCalendar(
            year=nepaliYYYY,
            month=nepaliMM,
            day_of_month=nepaliDD,
            day_of_week_in_month=(nepaliDD - 1) // 7 + 1,
            day_of_week=day_of_week,
            week_of_year=week_of_year,
            week_of_month=week_of_month,
            day_of_year=day_of_year,
            first_day_of_month=first_day_of_month,
            last_day_of_month=last_day_of_month,
            total_days_in_month=total_days_in_month,
            era=2
        )
        
    @staticmethod
    def convert_to_english_calendar(nepali_yyyy: int, nepali_mm: int, nepali_dd: int) -> CustomCalendar:
        if not DateConverters.is_nepali_calendar_in_conversion_range(nepali_yyyy, nepali_mm, nepali_dd):
            raise ValueError(f"Out of Range: Nepali year {nepali_yyyy} is out of range to convert.")

        starting_english_date, starting_nepali_calendar = DateConverters.initialize_starting_dates(nepali_yyyy, True)

        total_nep_days_count = DateConverters.calculate_total_nepali_days_count(
            starting_nepali_calendar, nepali_yyyy, nepali_mm, nepali_dd
        )

        english_yyyy, english_mm, english_dd = (
            starting_english_date.year,
            starting_english_date.month,
            starting_english_date.day_of_month,
        )

        day_of_week = starting_english_date.day_of_week
        day_of_year = starting_english_date.day_of_year
        week_of_year = starting_english_date.week_of_year
        week_of_month = starting_english_date.week_of_month

        first_day_of_month = starting_english_date.first_day_of_month
        last_day_of_month = starting_english_date.last_day_of_month

        total_days_in_month = DateConverters.get_total_days_in_english_month(english_yyyy, english_mm)

        for _ in range(total_nep_days_count):
            english_dd += 1
            day_of_week += 1
            day_of_year += 1

            if day_of_week > 7:
                day_of_week = 1
                week_of_year += 1
                week_of_month += 1

            if english_dd > total_days_in_month:
                english_mm += 1
                english_dd = 1
                week_of_month = 1

                if english_mm > 12:
                    english_yyyy += 1
                    english_mm = 1
                    day_of_year = 1
                    week_of_year = 1

                total_days_in_month = DateConverters.get_total_days_in_english_month(english_yyyy, english_mm)
                first_day_of_month = day_of_week

            remaining_days_of_the_month = total_days_in_month - english_dd
            last_day_of_month = (day_of_week + remaining_days_of_the_month) % 7

            if last_day_of_month == 0:
                last_day_of_month = 7

        return CustomCalendar(
            year=english_yyyy,
            month=english_mm,
            day_of_month=english_dd,
            first_day_of_month=first_day_of_month,
            last_day_of_month=last_day_of_month,
            day_of_week=day_of_week,
            day_of_week_in_month=(english_dd - 1) // 7 + 1,
            day_of_year=day_of_year,
            week_of_month=week_of_month,
            week_of_year=week_of_year,
            total_days_in_month=total_days_in_month,
            era=1,  # For English Date, the era is always 1 (for this library)
        )

     
    @staticmethod   
    def initialize_starting_dates(target_year: int, is_nepali_date: bool) -> tuple:
        reference_date = (
            nepali_date_map.get(target_year - 1)
            if is_nepali_date
            else english_date_map.get(target_year - 1)
        )

        starting_english_date = (
            reference_date.englishDate
            if reference_date and reference_date.englishDate
            else NepaliCalendarDefaults.startingEnglishCalendar
        )
        starting_nepali_calendar = (
            reference_date.nepaliDate
            if reference_date and reference_date.nepaliDate
            else NepaliCalendarDefaults.startingNepaliCalendar
        )

        return starting_english_date, starting_nepali_calendar

    @staticmethod
    def calculate_english_days_difference(starting_date: date, target_date: date) -> int:
        return (target_date - starting_date).days
    

    @staticmethod
    def calculate_total_nepali_days_count(
        starting_nepali_calendar: CustomCalendar, nepali_yyyy: int, nepali_mm: int, nepali_dd: int
    ) -> int:
        total_nep_days_count = 0

        for year in range(starting_nepali_calendar.year, nepali_yyyy):
            days_in_year = days_in_month_map[year]
            for month in range(1, 13):
                total_nep_days_count += days_in_year[month]

        for month in range(starting_nepali_calendar.month, nepali_mm):
            total_nep_days_count += days_in_month_map[nepali_yyyy][month]

        total_nep_days_count += nepali_dd - starting_nepali_calendar.day_of_month

        return total_nep_days_count

    @staticmethod
    def get_nepali_month(nepali_year: int, nepali_month: int, added_months_count: int) -> NepaliMonthCalendar:
        new_year, new_month = DateConverters.adjust_year_and_month(nepali_year, nepali_month + added_months_count)

        return DateConverters.calculate_nepali_month_details(new_year, new_month)

    @staticmethod
    def get_nepali_calendar(simple_nepali_date: SimpleDate) -> CustomCalendar:
        return DateConverters.get_custom_calendar_using_day_month_year(
            day_of_month=simple_nepali_date.day_of_month,
            month=simple_nepali_date.month,
            year=simple_nepali_date.year,
            adjust_month=False
        )
        
    # @staticmethod
    # def calculate_nepali_month_details(year: int, month: int) -> CustomCalendar:
    #     total_days_in_month = days_in_month_map.get(year, {}).get(month)
    #     print("My test mapper " + total_days_in_month)
    #     if total_days_in_month is None:
    #         raise ValueError(f"Invalid year {year} or month {month}.")

    #     starting_nepali_calendar = nepali_date_map.get(year - 1, None)
    #     print("My test mapper " + starting_nepali_calendar)
    #     if not starting_nepali_calendar:
    #         raise ValueError("Starting calendar not defined.")

    #     day_offset = DateConverters.calculate_day_offset(starting_nepali_calendar.year, year, month)
    #     first_day_of_month = (starting_nepali_calendar.first_day_of_month + day_offset) % 7
    #     normalized_first_day_of_month = 7 if first_day_of_month == 0 else first_day_of_month
    #     normalized_last_day_of_month = (
    #         (normalized_first_day_of_month + total_days_in_month - 1) % 7 or 7
    #     )

    #     return CustomCalendar(
    #         year=year,
    #         month=month,
    #         day_of_month=0,
    #         total_days_in_month=total_days_in_month,
    #         first_day_of_month=normalized_first_day_of_month,
    #         last_day_of_month=normalized_last_day_of_month,
    #         day_of_week_in_month=0,
    #         day_of_week=0,
    #         era=2,
    #         day_of_year=0,
    #         week_of_month=0,
    #         week_of_year=0,
    #     )

    @staticmethod
    def nepali_days_in_between(start_date: SimpleDate, end_date: SimpleDate) -> int:
        if start_date.year > end_date.year:
            return -DateConverters.nepali_days_in_between(end_date, start_date)

        if not (
            DateConverters.is_nepali_calendar_in_conversion_range(
                start_date.year, start_date.month, start_date.day_of_month
            )
            and DateConverters.is_nepali_calendar_in_conversion_range(
                end_date.year, end_date.month, end_date.day_of_month
            )
        ):
            raise ValueError(
                f"Out of range: Start year {start_date.year} or end year {end_date.year}."
            )

        start_offset = DateConverters.calculate_day_offset(start_date.year, start_date.year, start_date.month) + start_date.day_of_month
        end_offset = DateConverters.calculate_day_offset(start_date.year, end_date.year, end_date.month) + end_date.day_of_month

        return end_offset - start_offset

    @staticmethod
    def get_custom_calendar_using_day_month_year(year: int, month: int, day_of_month: int, adjust_month: bool) -> CustomCalendar:
        month_details = DateConverters.calculate_nepali_month_details(year, month)

        if adjust_month:
            new_day_of_month = min(day_of_month, month_details.total_days_in_month)
        else:
            if day_of_month > month_details.total_days_in_month:
                raise ValueError(f"Day {day_of_month} is out of range for month {month}.")
            new_day_of_month = day_of_month

        normalized_day_of_week = 7 if (month_details.first_day_of_month + new_day_of_month - 1) % 7 == 0 else (month_details.first_day_of_month + new_day_of_month - 1) % 7
        total_day_of_year = DateConverters.calculate_day_of_year(year, month, new_day_of_month)

        return CustomCalendar(
            year=year,
            month=month,
            day_of_month=new_day_of_month,
            total_days_in_month=month_details.total_days_in_month,
            first_day_of_month=month_details.first_day_of_month,
            last_day_of_month=month_details.last_day_of_month,
            day_of_week_in_month=(new_day_of_month - 1) // 7 + 1,
            day_of_week=normalized_day_of_week,
            era=2,
            day_of_year=total_day_of_year,
            week_of_month=DateConverters.calculate_week_of_month(
                day_of_month=day_of_month,
                first_day_of_month=month_details.first_day_of_month,
            ),
            week_of_year=DateConverters.calculate_week_of_year(
                total_day_of_year,
                DateConverters.calculate_nepali_month_details(year, 1).first_day_of_month,
            )
        )

    @staticmethod
    def adjust_nepali_date_for_day_adjustments(year: int, month: int, day_of_month: int, days_to_adjust: int) -> CustomCalendar:
        total_days_in_current_month = (
            days_in_month_map.get(year, [None])[month]
            if 1 <= month < len(days_in_month_map.get(year, [None]))
            else None
        )
        if total_days_in_current_month is None:
            raise ValueError(f"Invalid year {year} or month {month}.")

        if 1 <= day_of_month + days_to_adjust <= total_days_in_current_month:
            new_day = day_of_month + days_to_adjust
            return DateConverters.get_custom_calendar_using_day_month_year(year, month, new_day, True)

        if days_to_adjust > 0:
            remaining_days = days_to_adjust - (total_days_in_current_month - day_of_month + 1)
            new_month = month + 1
            new_year = year + 1 if new_month > 12 else year
            return DateConverters.adjust_nepali_date_for_day_adjustments(
                new_year, 1 if new_month > 12 else new_month, 1, remaining_days
            )
        else:
            new_month = month - 1
            new_year = year - 1 if new_month < 1 else year
            previous_month_days = days_in_month_map.get(new_year, [])[12 if new_month < 1 else new_month]

            if previous_month_days == 0:
                raise ValueError(f"Invalid year {new_year} or month {new_month}.")

            return DateConverters.adjust_nepali_date_for_day_adjustments(
                new_year,
                12 if new_month < 1 else new_month,
                previous_month_days,
                days_to_adjust + day_of_month,
            )
    
    @staticmethod
    def adjust_year_and_month(year: int, month: int) -> Tuple[int, int]:
        adjusted_year = year
        adjusted_month = month

        # Handle month overflow
        while adjusted_month > 12:
            adjusted_month -= 12
            adjusted_year += 1

        # Handle month underflow
        while adjusted_month < 1:
            adjusted_month += 12
            adjusted_year -= 1

        return adjusted_year, adjusted_month

    @staticmethod
    def calculate_nepali_month_details(nepali_year: int, nepali_month: int) -> NepaliMonthCalendar:

        if nepali_year not in days_in_month_map or nepali_month not in range(1, 13):
            raise ValueError(f"Invalid year {nepali_year} or month provided {nepali_month}.")
        
        total_days_in_month = days_in_month_map[nepali_year][nepali_month]

        starting_nepali_calendar = (
            nepali_date_map.get(nepali_year - 1, {}).nepaliDate 
            if nepali_date_map.get(nepali_year - 1) else NepaliCalendarDefaults.startingNepaliCalendar
        )
        
        day_offset = DateConverters.calculate_day_offset(starting_nepali_calendar.year, nepali_year, nepali_month)

        first_day_of_month = (starting_nepali_calendar.first_day_of_month + day_offset) % 7
        normalized_first_day_of_month = 7 if first_day_of_month == 0 else first_day_of_month

        normalized_last_day_of_month = ((normalized_first_day_of_month + total_days_in_month - 1) % 7)
        normalized_last_day_of_month = 7 if normalized_last_day_of_month == 0 else normalized_last_day_of_month

        return NepaliMonthCalendar(
            year=nepali_year,
            month=nepali_month,
            first_day_of_month=normalized_first_day_of_month,
            total_days_in_month=total_days_in_month,
            last_day_of_month=normalized_last_day_of_month
        )

    @staticmethod
    def calculate_day_offset(starting_year: int, target_year: int, target_month: int) -> int:        
        for year in range(starting_year, target_year):
            if year not in days_in_month_map:
                raise ValueError(f"Year {year} is missing in days_in_month_map.")
        
        year_offset = sum(
            sum(days_in_month_map[year][1:])
            for year in range(starting_year, target_year)
        )
        
        if target_year in days_in_month_map:
            month_offset = sum(days_in_month_map[target_year][1:target_month])
        else:
            raise ValueError(f"Target year {target_year} is missing in days_in_month_map.")
        
        return year_offset + month_offset

    @staticmethod
    def calculate_day_of_year(year: int, month: int, day_of_month: int) -> int:
        if year not in days_in_month_map:
            raise ValueError(f"Year {year} is not present in the days_in_month_map.")
        
        days_in_year = days_in_month_map[year]
        month_offset = sum(days_in_year[1:month])

        return month_offset + day_of_month

    @staticmethod
    def calculate_week_of_month(day_of_month: int, first_day_of_month: int) -> int:
        days_before = (day_of_month - 1) + (first_day_of_month - 1)
        return (days_before // 7) + 1

    @staticmethod
    def calculate_week_of_year(day_of_year: int, first_day_of_year: int) -> int:
        total_days_passed = day_of_year + (first_day_of_year - 1)
        return (total_days_passed // 7) + (1 if total_days_passed % 7 != 0 else 0)
    
    @staticmethod
    def get_total_days_in_english_month(year: int, month: int) -> int:
        if not (1 <= month <= 12):
            raise ValueError(f"Invalid month: {month}. Month must be between 1 and 12.")
        
        days_in_month = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        days_in_month_of_leap_year = [0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

        if DateConverters.is_english_leap_year(year):
            return days_in_month_of_leap_year[month]
        return days_in_month[month]

    @staticmethod
    def is_nepali_calendar_in_conversion_range(nepali_yyyy: int, nepali_mm: int, nepali_dd: int) -> bool:
        return (
            DateConverters.min_nepali_year <= nepali_yyyy <= DateConverters.max_nepali_year
            and 1 <= nepali_mm <= 12
            and 1 <= nepali_dd <= 32
        )
        
    @staticmethod
    def is_english_date_in_conversion_range(english_yyyy: int, english_mm: int, english_dd: int) -> bool:
        return (
            DateConverters.min_english_year <= english_yyyy <= DateConverters.max_english_year
            and 1 <= english_mm <= 12
            and 1 <= english_dd <= 31
        )

    @staticmethod
    def is_english_leap_year(year: int) -> bool:
        return (year % 4 == 0 and year % 100 != 0) or (year % 400 == 0)


import unittest
from nepali_calendar_utils.data.custom_calendar import *
from nepali_calendar_utils.calendar_model.nepali_calendar_model import NepaliCalendarModel
from nepali_calendar_utils.calendar_model.nepali_date_converter import NepaliDateConverter
from nepali_calendar_utils.data.nepali_date_locale import NameFormat, NepaliDateFormatStyle, NepaliDateLocale, NepaliCalendarUtilsLang

class TestNepaliDateConverter(unittest.TestCase):
    def setUp(self):
        self.calendar_model = NepaliCalendarModel()
        
    def test_nepali_to_english_date_converter_nepali_new_year_2079(self):
        converter = NepaliDateConverter()
        
        to_test_english_calendar = converter.convert_nepali_to_english(
            nepali_yyyy=2079, nepali_mm=1, nepali_dd=1
        )

        correct_english_custom_calendar = CustomCalendar(
            year=2022,
            month=4,
            day_of_month=14,
            era=1,
            first_day_of_month=6,
            last_day_of_month=7,
            total_days_in_month=30,
            day_of_week_in_month=2,
            day_of_week=5,
            day_of_year=104,
            week_of_month=3,
            week_of_year=16
        )

        self.assertEqual(correct_english_custom_calendar, to_test_english_calendar)

    def test_nepali_to_english_date_converter_nepali_new_year_2082(self):
        to_test_english_calendar = NepaliDateConverter.convert_nepali_to_english(
            nepali_yyyy=2082, nepali_mm=1, nepali_dd=1
        )

        correct_english_custom_calendar = CustomCalendar(
            year=2025,
            month=4,
            day_of_month=14,
            era=1,
            first_day_of_month=3,
            last_day_of_month=4,
            total_days_in_month=30,
            day_of_week_in_month=2,
            day_of_week=2,
            day_of_year=104,
            week_of_month=3,
            week_of_year=16
        )

        self.assertEqual(correct_english_custom_calendar, to_test_english_calendar)

    def test_nepali_to_english_date_converter_nepali_year_2100_11_12(self):
        to_test_english_calendar = NepaliDateConverter.convert_nepali_to_english(
            nepali_yyyy=2100, nepali_mm=11, nepali_dd=12
        )

        correct_english_custom_calendar = CustomCalendar(
            year=2044,
            month=2,
            day_of_month=24,
            era=1,
            first_day_of_month=2,
            last_day_of_month=2,
            total_days_in_month=29,
            day_of_week_in_month=4,
            day_of_week=4,
            day_of_year=55,
            week_of_month=4,
            week_of_year=9
        )

        self.assertEqual(correct_english_custom_calendar, to_test_english_calendar)

    def test_nepali_to_english_date_converter_nepali_year_2087_9_21(self):
        to_test_english_calendar = NepaliDateConverter.convert_nepali_to_english(
            nepali_yyyy=2087, nepali_mm=9, nepali_dd=21
        )

        correct_english_custom_calendar = CustomCalendar(
            year=2031,
            month=1,
            day_of_month=5,
            era=1,
            first_day_of_month=4,
            last_day_of_month=6,
            total_days_in_month=31,
            day_of_week_in_month=1,
            day_of_week=1,
            day_of_year=5,
            week_of_month=2,
            week_of_year=2
        )

        self.assertEqual(correct_english_custom_calendar, to_test_english_calendar)

    def test_nepali_to_english_date_converter_nepali_year_2097_3_17(self):
        to_test_english_calendar = NepaliDateConverter.convert_nepali_to_english(
            nepali_yyyy=2097, nepali_mm=3, nepali_dd=17
        )

        correct_english_custom_calendar = CustomCalendar(
            year=2040,
            month=7,
            day_of_month=1,
            era=1,
            first_day_of_month=1,
            last_day_of_month=3,
            total_days_in_month=31,
            day_of_week_in_month=1,
            day_of_week=1,
            day_of_year=183,
            week_of_month=1,
            week_of_year=27
        )

        self.assertEqual(correct_english_custom_calendar, to_test_english_calendar)

    def test_nepali_to_english_date_converter_nepali_year_2034_6_29(self):
        to_test_english_calendar = NepaliDateConverter.convert_nepali_to_english(
            nepali_yyyy=2034, nepali_mm=6, nepali_dd=29
        )

        correct_english_custom_calendar = CustomCalendar(
            year=1977,
            month=10,
            day_of_month=15,
            era=1,
            first_day_of_month=7,
            last_day_of_month=2,
            total_days_in_month=31,
            day_of_week_in_month=3,
            day_of_week=7,
            day_of_year=288,
            week_of_month=3,
            week_of_year=42
        )

        self.assertEqual(correct_english_custom_calendar, to_test_english_calendar)

    def test_english_to_nepali_date_converter_english_year_2000_6_24(self):
        to_test_nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
            english_yyyy=2000, english_mm=6, english_dd=24
        )

        correct_nepali_custom_calendar = CustomCalendar(
            year=2057,
            month=3,
            day_of_month=10,
            era=2,
            first_day_of_month=5,
            last_day_of_month=7,
            total_days_in_month=31,
            day_of_week_in_month=2,
            day_of_week=7,
            day_of_year=73,
            week_of_month=2,
            week_of_year=11
        )

        self.assertEqual(correct_nepali_custom_calendar, to_test_nepali_calendar)


    def test_english_to_nepali_date_converter_english_year_19770513(self):
        to_test_nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
            english_yyyy=1997, english_mm=5, english_dd=13
        )

        correct_nepali_custom_calendar = CustomCalendar(
            year=2054,
            month=1,
            day_of_month=31,
            era=2,
            first_day_of_month=1,
            last_day_of_month=3,
            total_days_in_month=31,
            day_of_week_in_month=5,
            day_of_week=3,
            day_of_year=31,
            week_of_month=5,
            week_of_year=5
        )

        self.assertEqual(correct_nepali_custom_calendar, to_test_nepali_calendar)

    def test_english_to_nepali_date_converter_english_year_19140808(self):
        to_test_nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
            english_yyyy=1914, english_mm=8, english_dd=8
        )

        correct_nepali_custom_calendar = CustomCalendar(
            year=1971,
            month=4,
            day_of_month=24,
            era=2,
            first_day_of_month=5,
            last_day_of_month=7,
            total_days_in_month=31,
            day_of_week_in_month=4,
            day_of_week=7,
            day_of_year=118,
            week_of_month=4,
            week_of_year=17
        )

        self.assertEqual(correct_nepali_custom_calendar, to_test_nepali_calendar)

    def test_english_to_nepali_date_converter_english_year_20250615(self):
        to_test_nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
            english_yyyy=2025, english_mm=6, english_dd=15
        )

        correct_nepali_custom_calendar = CustomCalendar(
            year=2082,
            month=3,
            day_of_month=1,
            era=2,
            first_day_of_month=1,
            last_day_of_month=4,
            total_days_in_month=32,
            day_of_week_in_month=1,
            day_of_week=1,
            day_of_year=63,
            week_of_month=1,
            week_of_year=10
        )

        self.assertEqual(correct_nepali_custom_calendar, to_test_nepali_calendar)

    def test_english_to_nepali_date_converter_english_leap_year_20240229(self):
        to_test_nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
            english_yyyy=2024, english_mm=2, english_dd=29
        )

        correct_nepali_custom_calendar = CustomCalendar(
            year=2080,
            month=11,
            day_of_month=17,
            era=2,
            first_day_of_month=3,
            last_day_of_month=4,
            total_days_in_month=30,
            day_of_week_in_month=3,
            day_of_week=5,
            day_of_year=322,
            week_of_month=3,
            week_of_year=47
        )

        self.assertEqual(correct_nepali_custom_calendar, to_test_nepali_calendar)

    def test_english_to_nepali_date_converter_english_year_20320605(self):
        to_test_nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
            english_yyyy=2032, english_mm=6, english_dd=5
        )

        correct_nepali_custom_calendar = CustomCalendar(
            year=2089,
            month=2,
            day_of_month=23,
            era=2,
            first_day_of_month=6,
            last_day_of_month=2,
            total_days_in_month=32,
            day_of_week_in_month=4,
            day_of_week=7,
            day_of_year=53,
            week_of_month=4,
            week_of_year=8
        )

        self.assertEqual(correct_nepali_custom_calendar, to_test_nepali_calendar)

    def test_english_to_nepali_date_converter_english_year_20431014_get_nepali_year_21000627(self):
        to_test_nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
            english_yyyy=2043, english_mm=10, english_dd=14
        )

        correct_nepali_custom_calendar = CustomCalendar(
            year=2100,
            month=6,
            day_of_month=27,
            era=2,
            first_day_of_month=6,
            last_day_of_month=7,
            total_days_in_month=30,
            day_of_week_in_month=4,
            day_of_week=4,
            day_of_year=184,
            week_of_month=5,
            week_of_year=27
        )

        self.assertEqual(correct_nepali_custom_calendar, to_test_nepali_calendar)

    # Test for a year limit of 2043 (commented out to indicate behavior)
    # def test_english_to_nepali_date_converter_english_year_20440412_get_nepali_year_21001230(self):
    #     to_test_nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
    #         english_yyyy=2044, english_mm=4, english_dd=12
    #     )
    #
    #     correct_nepali_custom_calendar = CustomCalendar(
    #         year=2100,
    #         month=12,
    #         day_of_month=30,
    #         era=2,
    #         first_day_of_month=2,
    #         last_day_of_month=4,
    #         total_days_in_month=31,
    #         day_of_week_in_month=5,
    #         day_of_week=3,
    #         day_of_year=365,
    #         week_of_month=5,
    #         week_of_year=53
    #     )
    #
    #     self.assertEqual(correct_nepali_custom_calendar, to_test_nepali_calendar)

    def test_english_to_nepali_date_converter_english_after_leap_year_20240302_get_nepali_year_20801119(self):
        to_test_nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
            english_yyyy=2024, english_mm=3, english_dd=2
        )

        correct_nepali_custom_calendar = CustomCalendar(
            year=2080,
            month=11,
            day_of_month=19,
            era=2,
            first_day_of_month=3,
            last_day_of_month=4,
            total_days_in_month=30,
            day_of_week_in_month=3,
            day_of_week=7,
            day_of_year=324,
            week_of_month=3,
            week_of_year=47
        )

        self.assertEqual(correct_nepali_custom_calendar, to_test_nepali_calendar)

    def test_english_to_nepali_date_converter_english_after_leap_year_first_day_of_week_20240303_get_nepali_year_20801120(self):
        to_test_nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
            english_yyyy=2024, english_mm=3, english_dd=3
        )

        correct_nepali_custom_calendar = CustomCalendar(
            year=2080,
            month=11,
            day_of_month=20,
            era=2,
            first_day_of_month=3,
            last_day_of_month=4,
            total_days_in_month=30,
            day_of_week_in_month=3,
            day_of_week=1,
            day_of_year=325,
            week_of_month=4,
            week_of_year=48
        )

        self.assertEqual(correct_nepali_custom_calendar, to_test_nepali_calendar)

    def test_date_conversion_convert_to_english_get_same_from_convert_to_nepali(self):
        nepali_calendar = NepaliDateConverter.convert_english_to_nepali(
            english_yyyy=2021, english_mm=2, english_dd=28
        )

        english_calendar = NepaliDateConverter.convert_nepali_to_english(
            nepali_yyyy=nepali_calendar.year,
            nepali_mm=nepali_calendar.month,
            nepali_dd=nepali_calendar.day_of_month
        )

        self.assertEqual(2021, english_calendar.year)
        self.assertEqual(2, english_calendar.month)
        self.assertEqual(28, english_calendar.day_of_month)
        

    def test_get_nepali_month_today_nepali_date_get_same_custom_month_and_nepali_calendar_properties(self):
        today = self.calendar_model.today_nepali_calendar

        nepali_month_from_simple_date = self.calendar_model.get_nepali_calendar(today.to_simple_date())
        nepali_month_from_year_and_month = self.calendar_model.get_nepali_month(
            nepali_year=today.year, nepali_month=today.month
        )

        self.assertEqual(today.to_nepali_month_calendar(), nepali_month_from_year_and_month)
        self.assertEqual(nepali_month_from_year_and_month, nepali_month_from_simple_date.to_nepali_month_calendar())

    def test_get_nepali_month_add_one_month_to_208203_get_nepali_calendar_of_date_208104(self):
        nepali_custom_calendar = CustomCalendar(
            year=2082, month=3, day_of_month=1, era=2,
            first_day_of_month=1, last_day_of_month=4, total_days_in_month=32,
            day_of_week_in_month=1, day_of_week=1, day_of_year=63,
            week_of_month=1, week_of_year=10
        )

        custom_calendar_after_one_month_added = self.calendar_model.plus_nepali_months(
            from_nepali_calendar=nepali_custom_calendar.to_nepali_month_calendar(), added_months_count=1
        )

        correct_custom_calendar_after_addition = CustomCalendar(
            year=2082, month=4, day_of_month=1, era=2,
            first_day_of_month=5, last_day_of_month=7, total_days_in_month=31,
            day_of_week_in_month=1, day_of_week=5, day_of_year=95,
            week_of_month=1, week_of_year=14
        )

        self.assertEqual(correct_custom_calendar_after_addition.year, custom_calendar_after_one_month_added.year)
        self.assertEqual(correct_custom_calendar_after_addition.month, custom_calendar_after_one_month_added.month)
        self.assertEqual(correct_custom_calendar_after_addition.first_day_of_month, custom_calendar_after_one_month_added.first_day_of_month)
        self.assertEqual(correct_custom_calendar_after_addition.last_day_of_month, custom_calendar_after_one_month_added.last_day_of_month)
        self.assertEqual(correct_custom_calendar_after_addition.total_days_in_month, custom_calendar_after_one_month_added.total_days_in_month)

    def test_calculate_positive_english_days_in_between(self):
        start_date = SimpleDate(2016, 2, 16)
        end_date = SimpleDate(2021, 3, 27)

        days_between = NepaliDateConverter.get_english_days_in_between(start_date, end_date)

        self.assertEqual(1866, days_between)

    def test_calculate_negative_english_days_in_between(self):
        start_date = SimpleDate(2024, 3, 8)
        end_date = SimpleDate(1980, 12, 31)

        days_between = NepaliDateConverter.get_english_days_in_between(start_date, end_date)

        self.assertEqual(-15773, days_between)

    def test_calculate_positive_nepali_days_in_between(self):
        start_date = SimpleDate(1980, 12, 31)
        end_date = SimpleDate(2081, 5, 24)

        days_between = NepaliDateConverter.get_nepali_days_in_between(start_date, end_date)

        self.assertEqual(36675, days_between)
        
        
    def test_get_nepali_month_today(self):
        today = self.calendar_model.today_nepali_calendar

        nepali_month_from_simple_date = self.calendar_model.get_nepali_calendar(
            simple_nepali_date=today.to_simple_date()
        )
        nepali_month_from_year_and_month = self.calendar_model.get_nepali_month(
            nepali_year=today.year, nepali_month=today.month
        )

        self.assertEqual(today.to_nepali_month_calendar(), nepali_month_from_year_and_month)
        self.assertEqual(
            nepali_month_from_year_and_month,
            nepali_month_from_simple_date.to_nepali_month_calendar()
        )

    def test_add_one_month(self):
        nepali_custom_calendar = CustomCalendar(
            year=2082, month=3, day_of_month=1, era=2,
            first_day_of_month=1, last_day_of_month=4,
            total_days_in_month=32, day_of_week_in_month=1,
            day_of_week=1, day_of_year=63, week_of_month=1,
            week_of_year=10
        )

        result = NepaliCalendarModel.plus_nepali_months(
            from_nepali_calendar=nepali_custom_calendar.to_nepali_month_calendar(),
            added_months_count=1
        )

        expected = CustomCalendar(
            year=2082, month=4, day_of_month=1, era=2,
            first_day_of_month=5, last_day_of_month=7,
            total_days_in_month=31, day_of_week_in_month=1,
            day_of_week=5, day_of_year=95, week_of_month=1,
            week_of_year=14
        )

        self.assertEqual(expected.year, result.year)
        self.assertEqual(expected.month, result.month)
        self.assertEqual(expected.first_day_of_month, result.first_day_of_month)
        self.assertEqual(expected.last_day_of_month, result.last_day_of_month)
        self.assertEqual(expected.total_days_in_month, result.total_days_in_month)

    def test_add_twenty_four_months(self):
        nepali_custom_calendar = CustomCalendar(
            year=2079, month=4, day_of_month=1, era=2,
            first_day_of_month=1, last_day_of_month=3,
            total_days_in_month=31, day_of_week_in_month=1,
            day_of_week=1, day_of_year=95, week_of_month=1,
            week_of_year=15
        )

        result = self.calendar_model.plus_nepali_months(
            from_nepali_calendar=nepali_custom_calendar.to_nepali_month_calendar(),
            added_months_count=24
        )

        expected = CustomCalendar(
            year=2081, month=4, day_of_month=1, era=2,
            first_day_of_month=3, last_day_of_month=6,
            total_days_in_month=32, day_of_week_in_month=1,
            day_of_week=3, day_of_year=95, week_of_month=1,
            week_of_year=15
        )

        self.assertEqual(expected.year, result.year)
        self.assertEqual(expected.month, result.month)
        self.assertEqual(expected.first_day_of_month, result.first_day_of_month)
        self.assertEqual(expected.last_day_of_month, result.last_day_of_month)
        self.assertEqual(expected.total_days_in_month, result.total_days_in_month)

    def test_subtract_four_months(self):
        nepali_custom_calendar = CustomCalendar(
            year=2082, month=4, day_of_month=1, era=2,
            first_day_of_month=5, last_day_of_month=7,
            total_days_in_month=31, day_of_week_in_month=1,
            day_of_week=5, day_of_year=95, week_of_month=1,
            week_of_year=14
        )

        result = self.calendar_model.minus_nepali_months(
            from_nepali_calendar=nepali_custom_calendar.to_nepali_month_calendar(),
            subtracted_months_count=4
        )

        expected = CustomCalendar(
            year=2081, month=12, day_of_month=1, era=2,
            first_day_of_month=6, last_day_of_month=1,
            total_days_in_month=31, day_of_week_in_month=1,
            day_of_week=6, day_of_year=336, week_of_month=1,
            week_of_year=49
        )

        self.assertEqual(expected.year, result.year)
        self.assertEqual(expected.month, result.month)
        self.assertEqual(expected.first_day_of_month, result.first_day_of_month)
        self.assertEqual(expected.last_day_of_month, result.last_day_of_month)
        self.assertEqual(expected.total_days_in_month, result.total_days_in_month)

    def test_calculate_first_and_last_day_of_month(self):
        custom_calendar_2079 = CustomCalendar(
            year=2079, month=4, day_of_month=1, era=2,
            first_day_of_month=1, last_day_of_month=3,
            total_days_in_month=31, day_of_week_in_month=1,
            day_of_week=1, day_of_year=95, week_of_month=1,
            week_of_year=15
        )

        custom_calendar_2082 = CustomCalendar(
            year=2082, month=4, day_of_month=1, era=2,
            first_day_of_month=5, last_day_of_month=7,
            total_days_in_month=31, day_of_week_in_month=1,
            day_of_week=5, day_of_year=95, week_of_month=1,
            week_of_year=14
        )

        result_2079 = NepaliDateConverter.get_nepali_month_calendar(
            nepali_year=custom_calendar_2079.year,
            nepali_month=custom_calendar_2079.month
        )
        result_2082 = NepaliDateConverter.get_nepali_month_calendar(
            nepali_year=custom_calendar_2082.year,
            nepali_month=custom_calendar_2082.month
        )

        self.assertEqual(custom_calendar_2079.year, result_2079.year)
        self.assertEqual(custom_calendar_2079.month, result_2079.month)
        self.assertEqual(custom_calendar_2079.first_day_of_month, result_2079.first_day_of_month)
        self.assertEqual(custom_calendar_2079.last_day_of_month, result_2079.last_day_of_month)
        self.assertEqual(custom_calendar_2079.total_days_in_month, result_2079.total_days_in_month)
        self.assertEqual(custom_calendar_2082.year, result_2082.year)
        self.assertEqual(custom_calendar_2082.month, result_2082.month)
        self.assertEqual(custom_calendar_2082.first_day_of_month, result_2082.first_day_of_month)
        self.assertEqual(custom_calendar_2082.last_day_of_month, result_2082.last_day_of_month)
        self.assertEqual(custom_calendar_2082.total_days_in_month, result_2082.total_days_in_month)


    def test_calculate_positive_english_days_in_between(self):
        start_date = SimpleDate(2016, 2, 16)
        end_date = SimpleDate(2021, 3, 27)

        days_between = NepaliDateConverter.get_english_days_in_between(start_date, end_date)

        self.assertEqual(days_between, 1866)

    def test_calculate_positive_english_days_in_between_v2(self):
        start_date = SimpleDate(1980, 12, 31)
        end_date = SimpleDate(2024, 3, 8)

        days_between = NepaliDateConverter.get_english_days_in_between(start_date, end_date)

        self.assertEqual(days_between, 15773)

    def test_calculate_negative_english_days_in_between(self):
        end_date = SimpleDate(1980, 12, 31)
        start_date = SimpleDate(2024, 3, 8)

        days_between = NepaliDateConverter.get_english_days_in_between(start_date, end_date)

        self.assertEqual(days_between, -15773)

    def test_calculate_positive_nepali_days_in_between(self):
        start_date = SimpleDate(1980, 12, 31)
        end_date = SimpleDate(2081, 5, 24)

        days_between = NepaliDateConverter.get_nepali_days_in_between(start_date, end_date)

        self.assertEqual(days_between, 36675)

    def test_calculate_negative_nepali_days_in_between(self):
        end_date = SimpleDate(1980, 12, 31)
        start_date = SimpleDate(2081, 5, 24)

        days_between = NepaliDateConverter.get_nepali_days_in_between(start_date, end_date)

        self.assertEqual(days_between, -36675)


    def test_compare_both_english_and_nepali_days_in_between(self):
        english_start_date = SimpleDate(1998, 4, 12)
        english_end_date = SimpleDate(2024, 9, 21)
        nepali_start_date = SimpleDate(2054, 12, 30)
        nepali_end_date = SimpleDate(2081, 6, 5)

        expected_days_difference = 9659

        nepali_days_between = NepaliDateConverter.get_nepali_days_in_between(nepali_start_date, nepali_end_date)
        english_days_between = NepaliDateConverter.get_english_days_in_between(english_start_date, english_end_date)

        self.assertEqual(expected_days_difference, nepali_days_between)
        self.assertEqual(expected_days_difference, english_days_between)
        self.assertEqual(nepali_days_between, english_days_between)

    def test_compare_calculated_nepali_calendar_using_day_month_year(self):
        converter = NepaliDateConverter()
        today = converter.today_nepali_calendar
        
        calculated_nepali_calendar_from_today = NepaliDateConverter.get_nepali_calendar(today.year, today.month, today.day_of_month)

        self.assertEqual(today, calculated_nepali_calendar_from_today)

        custom_start_calendar_of_2079 = CustomCalendar(
            year=2079,
            month=4,
            day_of_month=1,
            era=2,
            first_day_of_month=1,
            last_day_of_month=3,
            total_days_in_month=31,
            day_of_week_in_month=1,
            day_of_week=1,
            day_of_year=95,
            week_of_month=1,
            week_of_year=15
        )
        
        calculated_nepali_calendar_of_2079_start = NepaliDateConverter.get_nepali_calendar(
            custom_start_calendar_of_2079.year, custom_start_calendar_of_2079.month, custom_start_calendar_of_2079.day_of_month
        )

        self.assertEqual(custom_start_calendar_of_2079, calculated_nepali_calendar_of_2079_start)

        custom_start_calendar_of_2082 = CustomCalendar(
            year=2082,
            month=4,
            day_of_month=1,
            era=2,
            first_day_of_month=5,
            last_day_of_month=7,
            total_days_in_month=31,
            day_of_week_in_month=1,
            day_of_week=5,
            day_of_year=95,
            week_of_month=1,
            week_of_year=14
        )
        
        calculated_nepali_calendar_of_2082_start = NepaliDateConverter.get_nepali_calendar(
            custom_start_calendar_of_2082.year, custom_start_calendar_of_2082.month, custom_start_calendar_of_2082.day_of_month
        )

        self.assertEqual(custom_start_calendar_of_2082, calculated_nepali_calendar_of_2082_start)

        custom_end_calendar_of_2054 = CustomCalendar(
            year=2054,
            month=1,
            day_of_month=31,
            era=2,
            first_day_of_month=1,
            last_day_of_month=3,
            total_days_in_month=31,
            day_of_week_in_month=5,
            day_of_week=3,
            day_of_year=31,
            week_of_month=5,
            week_of_year=5
        )
        
        calculated_nepali_calendar_of_2054_end = NepaliDateConverter.get_nepali_calendar(
            custom_end_calendar_of_2054.year, custom_end_calendar_of_2054.month, custom_end_calendar_of_2054.day_of_month
        )

        self.assertEqual(custom_end_calendar_of_2054, calculated_nepali_calendar_of_2054_end)

    def test_compare_nepali_calendar_after_day_adjustments(self):
        converter = NepaliDateConverter()
        today = converter.today_nepali_calendar
        adjusted_calendar_of_today = NepaliDateConverter.get_nepali_calendar_after_addition_or_subtraction(
            today.year, today.month, today.day_of_month, 0
        )

        self.assertEqual(today, adjusted_calendar_of_today)

        custom_start_calendar_of_2082 = CustomCalendar(
            year=2082,
            month=4,
            day_of_month=1,
            era=2,
            first_day_of_month=5,
            last_day_of_month=7,
            total_days_in_month=31,
            day_of_week_in_month=1,
            day_of_week=5,
            day_of_year=95,
            week_of_month=1,
            week_of_year=14
        )

        adjusted_calendar_of_2082 = NepaliDateConverter.get_nepali_calendar_after_addition_or_subtraction(
            custom_start_calendar_of_2082.year, custom_start_calendar_of_2082.month, 20, -19
        )

        self.assertEqual(custom_start_calendar_of_2082, adjusted_calendar_of_2082)

        get_nepali_calendar_2081 = NepaliDateConverter.get_nepali_calendar(2081, 5, 28)
        adjusted_nepali_calendar_2081 = NepaliDateConverter.get_nepali_calendar_after_addition_or_subtraction(2081, 6, 9, -12)

        self.assertEqual(get_nepali_calendar_2081, adjusted_nepali_calendar_2081)

        get_nepali_calendar_2084 = NepaliDateConverter.get_nepali_calendar(2084, 1, 2)
        adjusted_nepali_calendar_2083 = NepaliDateConverter.get_nepali_calendar_after_addition_or_subtraction(2083, 12, 25, 7)

        self.assertEqual(get_nepali_calendar_2084, adjusted_nepali_calendar_2083)

    def test_compare_end_week_day_specifically_for_get_nepali_calendar(self):
        nepali_calendar_for_end_weekday_using_english_date = NepaliDateConverter.convert_english_to_nepali(2024, 9, 28)
        nepali_calendar_for_end_weekday_using_nepali_date = NepaliDateConverter.get_nepali_calendar(2081, 6, 12)

        nepali_calendar_for_end_year_end_weekday_using_english_date = NepaliDateConverter.convert_english_to_nepali(2025, 4, 12)
        nepali_calendar_for_end_year_end_weekday_using_nepali_date = NepaliDateConverter.get_nepali_calendar(2081, 12, 30)

        self.assertEqual(nepali_calendar_for_end_weekday_using_english_date, nepali_calendar_for_end_weekday_using_nepali_date)
        self.assertEqual(nepali_calendar_for_end_year_end_weekday_using_english_date, nepali_calendar_for_end_year_end_weekday_using_nepali_date)

    def test_compare_start_week_day_specifically_for_get_nepali_calendar(self):
        nepali_calendar_for_start_weekday_using_english_date = NepaliDateConverter.convert_english_to_nepali(2024, 9, 29)
        nepali_calendar_for_start_weekday_using_nepali_date = NepaliDateConverter.get_nepali_calendar(2081, 6, 13)

        nepali_calendar_for_end_year_start_weekday_using_english_date = NepaliDateConverter.convert_english_to_nepali(2025, 4, 13)
        nepali_calendar_for_end_year_start_weekday_using_nepali_date = NepaliDateConverter.get_nepali_calendar(2081, 12, 31)

        self.assertEqual(nepali_calendar_for_start_weekday_using_english_date, nepali_calendar_for_start_weekday_using_nepali_date)
        self.assertEqual(nepali_calendar_for_end_year_start_weekday_using_english_date, nepali_calendar_for_end_year_start_weekday_using_nepali_date)

    def test_format_nepali_date_random_date_get_formatted_date_in_string_by_locale(self):
        english_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.ENGLISH,
            date_format=NepaliDateFormatStyle.FULL,
            week_day_name=NameFormat.FULL,
            month_name=NameFormat.FULL
        )

        nepali_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.NEPALI,
            date_format=NepaliDateFormatStyle.MEDIUM,
            month_name=NameFormat.FULL
        )

        full_format_nepali_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.NEPALI,
            date_format=NepaliDateFormatStyle.FULL,
            week_day_name=NameFormat.FULL,
            month_name=NameFormat.FULL
        )

        to_test_formatted_date_in_english = NepaliDateConverter.format_nepali_date(
            year=2080, month=3, day_of_month=15, day_of_week=3, locale=english_locale
        )
        formatted_date_in_english = "Tuesday, Asar 15, 2080"

        to_test_formatted_date_in_nepali = NepaliDateConverter.format_nepali_date(
            year=2080, month=6, day_of_month=2, day_of_week=5, locale=nepali_locale
        )
        formatted_date_in_nepali = "२०८० असोज २"

        to_test_full_formatted_date_in_nepali = NepaliDateConverter.format_nepali_date(
            year=2080, month=2, day_of_month=2, day_of_week=5, locale=full_format_nepali_locale
        )
        full_formatted_date_in_nepali = "बिहिबार, जेठ २, २०८०"

        self.assertEqual(formatted_date_in_english, to_test_formatted_date_in_english)
        self.assertEqual(formatted_date_in_nepali, to_test_formatted_date_in_nepali)
        self.assertEqual(full_formatted_date_in_nepali, to_test_full_formatted_date_in_nepali)

    def test_format_nepali_date_using_other_helper_function_of_names_custom_calendar_date_get_formatted_date_in_string(self):
        date_converter = NepaliDateConverter()
        custom_calendar = date_converter.today_nepali_calendar

        week_day_name_in_english = NepaliDateConverter.get_weekday_name(
            day_of_week=custom_calendar.day_of_week, format=NameFormat.FULL, language=NepaliCalendarUtilsLang.ENGLISH
        )
        full_week_day_name_in_nepali = NepaliDateConverter.get_weekday_name(
            day_of_week=custom_calendar.day_of_week, format=NameFormat.FULL, language=NepaliCalendarUtilsLang.NEPALI
        )
        medium_week_day_name_in_nepali = NepaliDateConverter.get_weekday_name(
            day_of_week=custom_calendar.day_of_week, format=NameFormat.MEDIUM, language=NepaliCalendarUtilsLang.NEPALI
        )

        month_name_in_english = NepaliDateConverter.get_month_name(
            month=custom_calendar.month, format=NameFormat.FULL, language=NepaliCalendarUtilsLang.ENGLISH
        )
        month_name_in_nepali = NepaliDateConverter.get_month_name(
            month=custom_calendar.month, format=NameFormat.FULL, language=NepaliCalendarUtilsLang.NEPALI
        )
        short_month_name_in_nepali = NepaliDateConverter.get_month_name(
            month=custom_calendar.month, format=NameFormat.SHORT, language=NepaliCalendarUtilsLang.NEPALI
        )

        day_in_nepali = NepaliDateConverter.convert_to_nepali_number(str(custom_calendar.day_of_month))
        month_in_nepali = NepaliDateConverter.convert_to_nepali_number(str(custom_calendar.month))
        year_in_nepali = NepaliDateConverter.localize_number(str(custom_calendar.year), NepaliCalendarUtilsLang.NEPALI)

        english_locale = NepaliDateLocale()

        correct_formatted_date_in_english = f"{month_name_in_english} {custom_calendar.day_of_month}, {custom_calendar.year}"
        to_test_formatted_date_in_english = NepaliDateConverter.format_nepali_date_from_calendar(custom_calendar, english_locale)

        self.assertEqual(correct_formatted_date_in_english, to_test_formatted_date_in_english)

        full_nepali_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.NEPALI,
            date_format=NepaliDateFormatStyle.FULL
        )
        correct_full_nepali_formatted_date = f"{full_week_day_name_in_nepali}, {month_name_in_nepali} {day_in_nepali}, {year_in_nepali}"
        to_test_full_nepali_formatted_date = NepaliDateConverter.format_nepali_date_from_calendar(custom_calendar, full_nepali_locale)

        self.assertEqual(correct_full_nepali_formatted_date, to_test_full_nepali_formatted_date)

        mix_nepali_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.NEPALI,
            date_format=NepaliDateFormatStyle.FULL,
            week_day_name=NameFormat.MEDIUM,
            month_name=NameFormat.SHORT
        )
        correct_mix_nepali_formatted_date = f"{medium_week_day_name_in_nepali}, {short_month_name_in_nepali} {day_in_nepali}, {year_in_nepali}"
        to_test_mix_nepali_formatted_date = NepaliDateConverter.format_nepali_date_from_calendar(custom_calendar, mix_nepali_locale)

        self.assertEqual(correct_mix_nepali_formatted_date, to_test_mix_nepali_formatted_date)

        day_in_nepali_with_leading_zero = NepaliDateConverter.convert_to_nepali_number(str(custom_calendar.day_of_month).zfill(2))
        month_in_nepali_with_leading_zero = NepaliDateConverter.localize_number(str(custom_calendar.month).zfill(2), NepaliCalendarUtilsLang.NEPALI)

        compat_year = year_in_nepali[2:4]  # Alternatively, year_in_nepali[-2:]
        compat_nepali_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.NEPALI,
            date_format=NepaliDateFormatStyle.COMPACT_YMD
        )
        correct_compat_nepali_formatted_date = f"{compat_year}/{month_in_nepali_with_leading_zero}/{day_in_nepali_with_leading_zero}"
        to_test_compat_nepali_formatted_date = NepaliDateConverter.format_nepali_date_from_calendar(custom_calendar, compat_nepali_locale)

        self.assertEqual(correct_compat_nepali_formatted_date, to_test_compat_nepali_formatted_date)


    def test_format_english_date_random_date_get_formatted_date_in_string_by_locale(self):
        english_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.ENGLISH,
            date_format=NepaliDateFormatStyle.FULL,
            week_day_name=NameFormat.FULL,
            month_name=NameFormat.FULL
        )

        nepali_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.NEPALI,
            date_format=NepaliDateFormatStyle.MEDIUM,
            month_name=NameFormat.FULL
        )

        full_format_nepali_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.NEPALI,
            date_format=NepaliDateFormatStyle.FULL,
            week_day_name=NameFormat.FULL,
            month_name=NameFormat.FULL
        )

        to_test_formatted_date_in_english = NepaliDateConverter.format_english_date(
            year=2024, month=3, day_of_month=15, day_of_week=3, locale=english_locale
        )
        formatted_date_in_english = "Tuesday, March 15, 2024"

        to_test_formatted_date_in_nepali = NepaliDateConverter.format_english_date(
            year=2024, month=6, day_of_month=2, day_of_week=5, locale=nepali_locale
        )
        formatted_date_in_nepali = "२०२४ जुन २"

        to_test_full_formatted_date_in_nepali = NepaliDateConverter.format_english_date(
            year=2024, month=2, day_of_month=2, day_of_week=5, locale=full_format_nepali_locale
        )
        full_formatted_date_in_nepali = "बिहिबार, फेब्रुअरी २, २०२४"

        self.assertEqual(formatted_date_in_english, to_test_formatted_date_in_english)
        self.assertEqual(formatted_date_in_nepali, to_test_formatted_date_in_nepali)
        self.assertEqual(full_formatted_date_in_nepali, to_test_full_formatted_date_in_nepali)

    def test_format_english_date_using_other_helper_function_of_names_custom_calendar_date_get_formatted_date_in_string(self):
        date_converter = NepaliDateConverter()
        custom_calendar = date_converter.today_nepali_calendar
        today_nepali_date = date_converter.today_nepali_calendar
        custom_calendar = NepaliDateConverter.convert_nepali_to_english(
            today_nepali_date.year, today_nepali_date.month, today_nepali_date.day_of_month
        )

        week_day_name_in_english = NepaliDateConverter.get_weekday_name(
            day_of_week=custom_calendar.day_of_week, format=NameFormat.FULL, language=NepaliCalendarUtilsLang.ENGLISH
        )
        full_week_day_name_in_nepali = NepaliDateConverter.get_weekday_name(
            day_of_week=custom_calendar.day_of_week, format=NameFormat.FULL, language=NepaliCalendarUtilsLang.NEPALI
        )
        medium_week_day_name_in_nepali = NepaliDateConverter.get_weekday_name(
            day_of_week=custom_calendar.day_of_week, format=NameFormat.MEDIUM, language=NepaliCalendarUtilsLang.NEPALI
        )

        month_name_in_english = NepaliDateConverter.get_english_month_name(
            month=custom_calendar.month, format=NameFormat.FULL, language=NepaliCalendarUtilsLang.ENGLISH
        )
        month_name_in_nepali = NepaliDateConverter.get_english_month_name(
            month=custom_calendar.month, format=NameFormat.FULL, language=NepaliCalendarUtilsLang.NEPALI
        )
        short_month_name_in_nepali = NepaliDateConverter.get_english_month_name(
            month=custom_calendar.month, format=NameFormat.SHORT, language=NepaliCalendarUtilsLang.NEPALI
        )

        day_in_nepali = NepaliDateConverter.convert_to_nepali_number(str(custom_calendar.day_of_month))
        month_in_nepali = NepaliDateConverter.convert_to_nepali_number(str(custom_calendar.month))
        year_in_nepali = NepaliDateConverter.localize_number(str(custom_calendar.year), NepaliCalendarUtilsLang.NEPALI)

        english_locale = NepaliDateLocale()

        correct_formatted_date_in_english = f"{month_name_in_english} {custom_calendar.day_of_month}, {custom_calendar.year}"
        to_test_formatted_date_in_english = NepaliDateConverter.format_english_date_from_calendar(custom_calendar, english_locale)

        self.assertEqual(correct_formatted_date_in_english, to_test_formatted_date_in_english)

        full_nepali_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.NEPALI,
            date_format=NepaliDateFormatStyle.FULL
        )
        correct_full_nepali_formatted_date = f"{full_week_day_name_in_nepali}, {month_name_in_nepali} {day_in_nepali}, {year_in_nepali}"
        to_test_full_nepali_formatted_date = NepaliDateConverter.format_english_date_from_calendar(custom_calendar, full_nepali_locale)

        self.assertEqual(correct_full_nepali_formatted_date, to_test_full_nepali_formatted_date)

        mix_nepali_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.NEPALI,
            date_format=NepaliDateFormatStyle.FULL,
            week_day_name=NameFormat.MEDIUM,
            month_name=NameFormat.SHORT
        )
        correct_mix_nepali_formatted_date = f"{medium_week_day_name_in_nepali}, {short_month_name_in_nepali} {day_in_nepali}, {year_in_nepali}"
        to_test_mix_nepali_formatted_date = NepaliDateConverter.format_english_date_from_calendar(custom_calendar, mix_nepali_locale)

        self.assertEqual(correct_mix_nepali_formatted_date, to_test_mix_nepali_formatted_date)

        day_in_nepali_with_leading_zero = NepaliDateConverter.convert_to_nepali_number(str(custom_calendar.day_of_month).zfill(2))
        month_in_nepali_with_leading_zero = NepaliDateConverter.localize_number(str(custom_calendar.month).zfill(2), NepaliCalendarUtilsLang.NEPALI)

        compat_year = year_in_nepali[2:4]
        compat_nepali_locale = NepaliDateLocale(
            language=NepaliCalendarUtilsLang.NEPALI,
            date_format=NepaliDateFormatStyle.COMPACT_YMD
        )
        correct_compat_nepali_formatted_date = f"{compat_year}/{month_in_nepali_with_leading_zero}/{day_in_nepali_with_leading_zero}"
        to_test_compat_nepali_formatted_date = NepaliDateConverter.format_english_date_from_calendar(custom_calendar, compat_nepali_locale)

        self.assertEqual(correct_compat_nepali_formatted_date, to_test_compat_nepali_formatted_date)
        
    def test_format_time_in_english_random_simple_time(self):
        time = SimpleTime(16, 30, 48, 22)

        time_12_hour = NepaliDateConverter.get_formatted_time_in_english(time)
        correct_12_hour_formatted_time = "4:30 PM"
        self.assertEqual(correct_12_hour_formatted_time, time_12_hour)

        time_24_hour = NepaliDateConverter.get_formatted_time_in_english(time, False)
        correct_24_hour_formatted_time = "16:30"
        self.assertEqual(correct_24_hour_formatted_time, time_24_hour)

    def test_format_time_in_nepali_random_simple_time(self):
        time = SimpleTime(16, 30, 48, 22)

        time_12_hour = NepaliDateConverter.get_formatted_time_in_nepali(time)
        correct_12_hour_formatted_time = "दिउँसो ४ : ३०"
        self.assertEqual(correct_12_hour_formatted_time, time_12_hour)

        time_24_hour = NepaliDateConverter.get_formatted_time_in_nepali(time, False)
        correct_24_hour_formatted_time = "१६ : ३०"
        self.assertEqual(correct_24_hour_formatted_time, time_24_hour)

    def test_format_time_to_iso_format_random_dates(self):
            nepali_date = SimpleDate(2081, 5, 24)
            time = SimpleTime(14, 45, 15, 123456789)

            nepali_date_iso_format = NepaliDateConverter.format_nepali_datetime_to_iso(nepali_date, time)
            correct_nepali_date_iso_format = "2024-09-09T09:00:15.123456Z"
            self.assertEqual(correct_nepali_date_iso_format, nepali_date_iso_format)

            english_date = SimpleDate(2024, 9, 9)
            english_date_iso_format = NepaliDateConverter.format_english_date_nepali_time_to_iso(english_date, time)
            correct_english_date_iso_format = "2024-09-09T09:00:15.123456Z"
            self.assertEqual(correct_english_date_iso_format, english_date_iso_format)

            self.assertEqual(nepali_date_iso_format, english_date_iso_format)

    def test_format_and_compare_time_to_iso_format_today_english_and_nepali_date(self):
        date_convert = NepaliDateConverter()
        time = SimpleTime(14, 30, 15, 0)

        nepali_date = date_convert.today_nepali_calendar.to_simple_date()
        nepali_date_iso_format = NepaliDateConverter.format_nepali_datetime_to_iso(nepali_date, time)

        english_date = date_convert.today_english_simple_date
        english_date_iso_format = NepaliDateConverter.format_english_date_nepali_time_to_iso(english_date, time)

        self.assertEqual(nepali_date_iso_format, english_date_iso_format)
        
    def test_convert_iso_to_nepali_calendar_and_time(self):
        iso_string = "2024-09-09T09:00:15Z"
        custom_datetime = NepaliDateConverter.get_nepali_date_time_from_iso_format(iso_string)
        expected_calendar = NepaliDateConverter.get_nepali_calendar(2081, 5, 24)
        expected_time = SimpleTime(14, 45, 15, 0)

        self.assertEqual(expected_calendar, custom_datetime.custom_calendar)
        self.assertEqual(expected_time, custom_datetime.simple_time)

        iso_string2 = "2020-08-30T18:43:00.123456Z"
        custom_datetime2 = NepaliDateConverter.get_nepali_date_time_from_iso_format(iso_string2)
        expected_calendar2 = NepaliDateConverter.get_nepali_calendar(2077, 5, 15)
        expected_time2 = SimpleTime(0, 28, 0, 123456000)

        self.assertEqual(expected_calendar2, custom_datetime2.custom_calendar)
        self.assertEqual(expected_time2, custom_datetime2.simple_time)
        
    def test_various_iso_patterns_to_nepali_calendar_and_time(self):
        test_cases = [
            ("2020-08-30T18:43:00Z", (2077, 5, 15), SimpleTime(0, 28, 0, 0)),
            ("2020-08-30T18:43:00.502Z", (2077, 5, 15), SimpleTime(0, 28, 0, 502000000)),
            ("2020-08-30T18:43:00.123456Z", (2077, 5, 15), SimpleTime(0, 28, 0, 123456000)),
            ("2020-08-30T18:40:00+00:00", (2077, 5, 15), SimpleTime(0, 25, 0, 0)),
            ("2020-08-30T18:40:00+00:00:00", (2077, 5, 15), SimpleTime(0, 25, 0, 0)),
            ("2011-11-04", (2068, 7, 18), SimpleTime(0, 0, 0, 0)),
            ("2011-11-04 00:05:23.283", (2068, 7, 18), SimpleTime(0, 5, 23, 283000000)),
            ("2011-11-04 00:05:23.283+00:00", (2068, 7, 18), SimpleTime(5, 50, 23, nanosecond=283000000)),
            ("2011-11-04T00:05:23+04:00", (2068, 7, 18), SimpleTime(1, 50, 23, 0)),
        ]

        for iso_string, (ny, nm, nd), expected_time in test_cases:
            with self.subTest(iso_string=iso_string):
                result = NepaliDateConverter.get_nepali_date_time_from_iso_format(iso_string)
                expected_calendar = NepaliDateConverter.get_nepali_calendar(ny, nm, nd)
                self.assertEqual(result.custom_calendar, expected_calendar)
                self.assertEqual(result.simple_time, expected_time)


    def test_convert_iso_to_english_calendar_and_nepali_time(self):
        iso_string = "2024-09-09T09:00:15Z"
        custom_datetime = NepaliDateConverter.get_english_date_nepali_time_from_iso_format(iso_string)

        expected_nepali_calendar = NepaliDateConverter.get_nepali_calendar(2081, 5, 24)
        expected_english_calendar = NepaliDateConverter.convert_nepali_to_english(
            expected_nepali_calendar.year,
            expected_nepali_calendar.month,
            expected_nepali_calendar.day_of_month
        )
        expected_time = SimpleTime(14, 45, 15, 0)

        self.assertEqual(expected_english_calendar, custom_datetime.custom_calendar)
        self.assertEqual(expected_time, custom_datetime.simple_time)

        iso_string2 = "2020-01-01T23:59:59.123456Z"
        custom_datetime2 = NepaliDateConverter.get_english_date_nepali_time_from_iso_format(iso_string2)

        expected_nepali_calendar2 = NepaliDateConverter.get_nepali_calendar(2076, 9, 17)
        expected_english_calendar2 = NepaliDateConverter.convert_nepali_to_english(
            expected_nepali_calendar2.year,
            expected_nepali_calendar2.month,
            expected_nepali_calendar2.day_of_month
        )
        expected_time2 = SimpleTime(5, 44, 59, 123456000)

        self.assertEqual(expected_english_calendar2, custom_datetime2.custom_calendar)
        self.assertEqual(expected_time2, custom_datetime2.simple_time)
        
    def test_various_iso_patterns_to_english_calendar_and_nepali_time(self):
        test_cases = [
            ("2020-08-30T18:43:00Z", (2077, 5, 15), SimpleTime(0, 28, 0, 0)),
            ("2020-08-30T18:43:00.502Z", (2077, 5, 15), SimpleTime(0, 28, 0, 502000000)),
            ("2020-08-30T18:43:00.123456Z", (2077, 5, 15), SimpleTime(0, 28, 0, 123456000)),
            ("2020-08-30T18:40:00+00:00", (2077, 5, 15), SimpleTime(0, 25, 0, 0)),
            ("2020-08-30T18:40:00+00:00:00", (2077, 5, 15), SimpleTime(0, 25, 0, 0)),
            ("2011-11-04", (2068, 7, 18), SimpleTime(0, 0, 0, 0)),
            ("2011-11-04 00:05:23.283", (2068, 7, 18), SimpleTime(0, 5, 23, 283000000)),
            ("2011-11-04 00:05:23.283+00:00", (2068, 7, 18), SimpleTime(5, 50, 23, 283000000)),
            ("2011-11-04T00:05:23+04:00", (2068, 7, 18), SimpleTime(1, 50, 23, 0)),
        ]

        for iso_string, (ny, nm, nd), expected_time in test_cases:
            with self.subTest(iso_string=iso_string):
                custom_datetime = NepaliDateConverter.get_english_date_nepali_time_from_iso_format(iso_string)
                nepali_calendar = NepaliDateConverter.get_nepali_calendar(ny, nm, nd)
                expected_english_calendar = NepaliDateConverter.convert_nepali_to_english(
                    nepali_calendar.year, nepali_calendar.month, nepali_calendar.day_of_month
                )
                self.assertEqual(custom_datetime.custom_calendar, expected_english_calendar)
                self.assertEqual(custom_datetime.simple_time, expected_time)

        
    def test_replace_delimiters_of_string_random_date_string_with_delimiter(self):
        original_date = "2024/06/21"
        new_delimiter = "-"
        formatted_date = NepaliDateConverter.replace_delimiter(original_date, new_delimiter)
        correct_formatted_date = "2024-06-21"
        self.assertEqual(correct_formatted_date, formatted_date)

        original_nepali_date = "२०२४/०६/२१"
        formatted_nepali_date = NepaliDateConverter.replace_delimiter(original_nepali_date, new_delimiter)
        correct_formatted_nepali_date = "२०२४-०६-२१"
        self.assertEqual(correct_formatted_nepali_date, formatted_nepali_date)

        original_time = "09:45 AM"
        new_delimiter_space = " "
        old_delimiter = ":"
        formatted_time_with_space = NepaliDateConverter.replace_delimiter(original_time, new_delimiter_space, old_delimiter)
        corrected_formatted_time_with_space = "09 45 AM"
        self.assertEqual(corrected_formatted_time_with_space, formatted_time_with_space)

    
    def test_format_nepali_datetime_full_pattern_returns_correct_format(self):
        test_calendar = NepaliDateConverter.get_nepali_calendar(2081, 5, 24)
        test_time = SimpleTime(14, 45, 15, 123000000)

        def check(pattern, expected, lang=NepaliCalendarUtilsLang.NEPALI):
            result = NepaliDateConverter.format_nepali_date_time_by_unicode_pattern(pattern, test_calendar, test_time, lang)
            self.assertEqual(expected, result, f"Pattern: {pattern}")

        check("yyyy", "२०८१")
        check("yy", "८१")
        check("MMMM", "भदौ")
        check("MMM", "भ")
        check("MM", "०५")
        check("M", "५")
        check("dd", "२४")
        check("d", "२४")
        check("D", "१५०")
        check("w", "२३")

        check("EEEE", "Monday", lang=NepaliCalendarUtilsLang.ENGLISH)
        check("EEEE", "सोमबार")
        check("E", "सोम")
        check("EEEEE", "सो")

        check("ee", "०२")
        check("e", "२")

        check("HH", "१४")
        check("H", "१४")
        check("hh", "०२")
        check("h", "२")
        check("mm", "४५")
        check("m", "४५")
        check("ss", "१५")
        check("s", "१५")

        check("SSSS", "१२३०")
        check("SSS", "१२३")
        check("SS", "१२")
        check("S", "१")

        check("a", "दिउँसो")
        check("A", "दिउँसो")

        check("yyyy-MM-dd E a hh:mm:ss", "२०८१-०५-२४ सोम दिउँसो ०२:४५:१५")

    def test_format_english_datetime_full_pattern_returns_correct_format(self):
        test_calendar = NepaliDateConverter.get_nepali_calendar(2081, 5, 24)
        test_time = SimpleTime(14, 45, 15, 123000000)

        def check(pattern, expected):
            result = NepaliDateConverter.format_english_date_time_by_unicode_pattern(pattern, test_calendar, test_time)
            self.assertEqual(expected, result, f"Pattern: {pattern}")

        check("yyyy", "2081")
        check("yy", "81")
        check("MMMM", "May")
        check("MMM", "May")
        check("MM", "05")
        check("M", "5")
        check("dd", "24")
        check("d", "24")
        check("D", "150")
        check("w", "23")

        check("EEEE", "Monday")
        check("E", "Mon")
        check("EEEEE", "M")

        check("ee", "02")
        check("e", "2")

        check("HH", "14")
        check("H", "14")
        check("hh", "02")
        check("h", "2")
        check("mm", "45")
        check("m", "45")
        check("ss", "15")
        check("s", "15")

        check("SSSS", "1230")
        check("SSS", "123")
        check("SS", "12")
        check("S", "1")

        check("a", "pm")
        check("A", "PM")

        check("yyyy-MM-dd E A hh:mm:ss", "2081-05-24 Mon PM 02:45:15")

    def test_format_datetime_midnight_and_end_of_day_correct_format(self):
        calendar = NepaliDateConverter.get_nepali_calendar(2081, 1, 1)

        midnight = SimpleTime(0, 0, 0, 0)
        end_of_day = SimpleTime(23, 59, 59, 999000000)

        def check(time, expected):
            result = NepaliDateConverter.format_english_date_time_by_unicode_pattern("HH:mm:ss.SSS a", calendar, time)
            self.assertEqual(expected, result)

        check(midnight, "00:00:00.000 am")
        check(end_of_day, "23:59:59.999 pm")

    def test_format_datetime_empty_pattern_returns_empty_string(self):
        calendar = NepaliDateConverter.get_nepali_calendar(2081, 1, 1)
        result = NepaliDateConverter.format_english_date_time_by_unicode_pattern("", calendar)
        self.assertEqual("", result)

    def test_format_datetime_unsupported_tokens_ignores_them(self):
        calendar = NepaliDateConverter.get_nepali_calendar(2081, 1, 1)
        pattern = "yyyy-MM-XX HH:mm:ss ??"
        result = NepaliDateConverter.format_english_date_time_by_unicode_pattern(pattern, calendar, SimpleTime(10, 30, 0, 0))
        self.assertEqual("2081-01-XX 10:30:00 ??", result)

    def test_format_datetime_all_tokens_mixed_returns_complete_format(self):
        calendar = NepaliDateConverter.get_nepali_calendar(2081, 5, 24)
        time = SimpleTime(14, 45, 15, 123000000)
        pattern = "yyyy yy MMMM MMM MM M dd d D EEEE E EEEEE ee e w HH H hh h mm m ss s SSSS SSS SS S a A"
        expected = "2081 81 May May 05 5 24 24 150 Monday Mon M 02 2 23 14 14 02 2 45 45 15 15 1230 123 12 1 pm PM"

        result = NepaliDateConverter.format_english_date_time_by_unicode_pattern(pattern, calendar, time)
        self.assertEqual(expected, result)

    def test_format_nepali_datetime_all_tokens_mixed_returns_complete_format(self):
        calendar = NepaliDateConverter.get_nepali_calendar(2081, 5, 24)
        time = SimpleTime(14, 45, 15, 123000000)
        pattern = "yyyy yy MMMM MMM MM M dd d D EEEE E EEEEE ee e w HH H hh h mm m ss s SSSS SSS SS S a A"
        expected = "२०८१ ८१ भदौ भ ०५ ५ २४ २४ १५० सोमबार सोम सो ०२ २ २३ १४ १४ ०२ २ ४५ ४५ १५ १५ १२३० १२३ १२ १ दिउँसो दिउँसो"

        result = NepaliDateConverter.format_nepali_date_time_by_unicode_pattern(pattern, calendar, time)
        self.assertEqual(expected, result)

    def test_format_only_date_english_and_nepali(self):
        calendar = NepaliDateConverter.get_nepali_calendar(2081, 5, 24)
        en = NepaliDateConverter.format_english_date_time_by_unicode_pattern("yyyy-MM-dd", calendar)
        np = NepaliDateConverter.format_nepali_date_time_by_unicode_pattern("yyyy-MM-dd", calendar, language=NepaliCalendarUtilsLang.NEPALI)

        self.assertEqual("2081-05-24", en)
        self.assertEqual("२०८१-०५-२४", np)

    def test_format_only_time_am_pm_language_sensitive(self):
        morning = SimpleTime(9, 5, 3, 45000000)
        night = SimpleTime(22, 30, 0, 0)

        self.assertEqual("09:05:03 am", NepaliDateConverter.format_time_by_unicode_pattern("hh:mm:ss a", morning, NepaliCalendarUtilsLang.ENGLISH))
        self.assertEqual("१०:३०:०० राति", NepaliDateConverter.format_time_by_unicode_pattern("hh:mm:ss a", night, NepaliCalendarUtilsLang.NEPALI))

    def test_format_ampm_language_difference(self):
        morning = SimpleTime(5, 0, 0, 0)
        evening = SimpleTime(18, 0, 0, 0)

        self.assertEqual("am", NepaliDateConverter.format_time_by_unicode_pattern("a", morning, NepaliCalendarUtilsLang.ENGLISH))
        self.assertEqual("pm", NepaliDateConverter.format_time_by_unicode_pattern("a", evening, NepaliCalendarUtilsLang.ENGLISH))

        self.assertEqual("बिहान", NepaliDateConverter.format_time_by_unicode_pattern("a", morning, NepaliCalendarUtilsLang.NEPALI))
        self.assertEqual("साँझ", NepaliDateConverter.format_time_by_unicode_pattern("a", evening, NepaliCalendarUtilsLang.NEPALI))

        self.assertEqual("AM", NepaliDateConverter.format_time_by_unicode_pattern("A", morning, NepaliCalendarUtilsLang.ENGLISH))
        self.assertEqual("PM", NepaliDateConverter.format_time_by_unicode_pattern("A", evening, NepaliCalendarUtilsLang.ENGLISH))

        self.assertEqual("बिहान", NepaliDateConverter.format_time_by_unicode_pattern("A", morning, NepaliCalendarUtilsLang.NEPALI))
        self.assertEqual("साँझ", NepaliDateConverter.format_time_by_unicode_pattern("A", evening, NepaliCalendarUtilsLang.NEPALI))

if __name__ == "__main__":
    unittest.main()


from dataclasses import dataclass
from enum import Enum

class NameFormat(Enum):
    FULL = "full"
    MEDIUM = "medium"
    SHORT = "short"

class NepaliDateFormatStyle(Enum):
    FULL = "full"               # Monday, Asar 21, 2024 or सोमबार, असार २१, २०२४
    LONG = "long"               # Asar 21, 2024 or असार २१, २०२४
    MEDIUM = "medium"           # 2024 Asar 21 or २०२४ असार २१
    SHORT_MDY = "short_mdy"     # 06/21/2024 or ०६/२१/२०२४
    SHORT_YMD = "short_ymd"     # 2024/06/21 or २०२४/०६/२१
    COMPACT_MDY = "compact_mdy" # 06/21/24 or ०६/२१/२४
    COMPACT_YMD = "compact_ymd" # 24/06/21 or २४/०६/२१

@dataclass(frozen=True)
class NepaliWeekdayName:
    short: str
    medium: str
    full: str

@dataclass(frozen=True)
class NepaliMonthName:
    short: str
    full: str

nepali_months = [
    NepaliMonthName(short="बै", full="बैशाख"),
    NepaliMonthName(short="जे", full="जेठ"),
    NepaliMonthName(short="अ", full="असार"),
    NepaliMonthName(short="सा", full="साउन"),
    NepaliMonthName(short="भ", full="भदौ"),
    NepaliMonthName(short="अ", full="असोज"),
    NepaliMonthName(short="का", full="कार्तिक"),
    NepaliMonthName(short="मं", full="मंसिर"),
    NepaliMonthName(short="पु", full="पौष"),
    NepaliMonthName(short="मा", full="माघ"),
    NepaliMonthName(short="फा", full="फाल्गुन"),
    NepaliMonthName(short="चै", full="चैत"),
]

nepali_months_in_english = [
    NepaliMonthName(short="Bai", full="Baisakh"),
    NepaliMonthName(short="Jes", full="Jestha"),
    NepaliMonthName(short="Asa", full="Asar"),
    NepaliMonthName(short="Shr", full="Shrawn"),
    NepaliMonthName(short="Bha", full="Bhadra"),
    NepaliMonthName(short="Aso", full="Asoj"),
    NepaliMonthName(short="Kar", full="Kartik"),
    NepaliMonthName(short="Man", full="Mangsir"),
    NepaliMonthName(short="Pou", full="Poush"),
    NepaliMonthName(short="Mag", full="Magh"),
    NepaliMonthName(short="Pha", full="Falgun"),
    NepaliMonthName(short="Chai", full="Chaitra"),
]

nepali_weekdays = [
    NepaliWeekdayName(short="आ", medium="आईत", full="आईतबार"),
    NepaliWeekdayName(short="सो", medium="सोम", full="सोमबार"),
    NepaliWeekdayName(short="मं", medium="मंगल", full="मंगलबार"),
    NepaliWeekdayName(short="बु", medium="बुध", full="बुधबार"),
    NepaliWeekdayName(short="बि", medium="बिहि", full="बिहिबार"),
    NepaliWeekdayName(short="शु", medium="शुक्र", full="शुक्रबार"),
    NepaliWeekdayName(short="श", medium="शनि", full="शनिबार"),
]

english_weekdays = [
    NepaliWeekdayName(short="S", medium="Sun", full="Sunday"),
    NepaliWeekdayName(short="M", medium="Mon", full="Monday"),
    NepaliWeekdayName(short="T", medium="Tue", full="Tuesday"),
    NepaliWeekdayName(short="W", medium="Wed", full="Wednesday"),
    NepaliWeekdayName(short="T", medium="Thu", full="Thursday"),
    NepaliWeekdayName(short="F", medium="Fri", full="Friday"),
    NepaliWeekdayName(short="S", medium="Sat", full="Saturday"),
]

english_months_in_english = [
    NepaliMonthName(short="Jan", full="January"),
    NepaliMonthName(short="Feb", full="February"),
    NepaliMonthName(short="Mar", full="March"),
    NepaliMonthName(short="Apr", full="April"),
    NepaliMonthName(short="May", full="May"),
    NepaliMonthName(short="Jun", full="June"),
    NepaliMonthName(short="Jul", full="July"),
    NepaliMonthName(short="Aug", full="August"),
    NepaliMonthName(short="Sep", full="September"),
    NepaliMonthName(short="Oct", full="October"),
    NepaliMonthName(short="Nov", full="November"),
    NepaliMonthName(short="Dec", full="December"),
]

english_months_in_nepali = [
    NepaliMonthName(short="जन", full="जनवरी"),
    NepaliMonthName(short="फेब्रु", full="फेब्रुअरी"),
    NepaliMonthName(short="मार्च", full="मार्च"),
    NepaliMonthName(short="अप्रि", full="अप्रिल"),
    NepaliMonthName(short="मे", full="मे"),
    NepaliMonthName(short="जुन", full="जुन"),
    NepaliMonthName(short="जुला", full="जुलाई"),
    NepaliMonthName(short="अग", full="अगस्ट"),
    NepaliMonthName(short="सेप्ट", full="सेप्टेम्बर"),
    NepaliMonthName(short="अक्टो", full="अक्टोबर"),
    NepaliMonthName(short="नोभे", full="नोभेम्बर"),
    NepaliMonthName(short="डिसे", full="डिसेम्बर"),
]

class NepaliCalendarUtilsLang(Enum):
    ENGLISH = {
        "weekdays": english_weekdays,
        "months": nepali_months_in_english,
        "english_months": english_months_in_english
    }

    NEPALI = {
        "weekdays": nepali_weekdays,
        "months": nepali_months,
        "english_months": english_months_in_nepali
    }

    def __init__(self, values):
        self.weekdays = values["weekdays"]
        self.months = values["months"]
        self.english_months = values["english_months"]
        
@dataclass(frozen=True)
class NepaliDateLocale:
    language: NepaliCalendarUtilsLang = NepaliCalendarUtilsLang.ENGLISH
    date_format: NepaliDateFormatStyle = NepaliDateFormatStyle.LONG
    week_day_name: NameFormat = NameFormat.FULL
    month_name: NameFormat = NameFormat.FULL
